/* Portwatch, by Mark Kamichoff [kamicm@rpi.edu] *
 * Original Filename: portwatch.c                *
 * Opens a TCP port and displays attempted       *
 * connections to it.                            */

#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <signal.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <sys/wait.h>

#include "logmanip.h"

#define VERSION_NO "0.1pre2"

#define _UTS_NAMESIZE 16
#define _UTS_NODESIZE 256

#define SA struct sockaddr	/* Richard Stevens suggests this
				   and it makes life easier! */

int listeningfd, clientfd;	/* Global socket file descriptors */

/* Sets up signal prototypes */

void signalpipe(void);		/* if we write something that's gone */
void signalterminate(void);	/* SIGTERM */
void signalinterrupt(void);	/* SIGINT (^C) */
void signalchild(void);		/* page 128 */
void setupsignals(void);	/* sets up signal handlers --^ */

void signalpipe(void) {
	fputs("caught SIGPIPE...broken file descriptor; exiting...\n",
		stderr);
	close(listeningfd);
	close(clientfd);
	fputs("closed client and server file descriptors...\n",stderr);
	exit(1);
}

void signalterminate(void) {
	fputs("caught SIGTERM...exiting now...\n",stderr);
	close(listeningfd);
	close(clientfd);
	fputs("closed client and server file descriptors...\n",stderr);
	exit(1);
}

void signalinterrupt(void) {
	fputs("caught SIGINT...exiting now...\n",stderr);
	close(listeningfd);
	close(clientfd);
	fputs("closed client and server file descriptors...\n",stderr);
	exit(1);
}

void signalchild(void)
{
	pid_t pid;
	int stat;

	while((pid = waitpid(-1, &stat, WNOHANG)) > 0);
//		printf("child %d terminated\n", pid);

	/* Change to something more informative
	   and relating to our current situation */

	return;
}

void setupsignals(void)
{
	/* Setup signal handlers */

        signal(SIGPIPE,(void *) signalpipe);
        signal(SIGINT,(void *) signalinterrupt);
        signal(SIGTERM,(void *) signalterminate);
	signal(SIGCHLD,(void *) signalchild);
}

int main(int argc, char **argv)
{
	socklen_t clientlength;
	struct sockaddr_in servaddr, clientaddress;
	struct hostent *hptr;
	struct utsname myname;
	time_t ticks;
	char logging = 0;
	char junkbuffer[MAXSIZE];
	int n;	/* For UNIX read() */

	setupsignals();	/* Sets up signal handlers */

	if( (argc > 1) && !strcmp(argv[1],"--version") )
	{
		printf("%s\n\n",VERSION_NO);
		exit(0);
	}

	printf("--> Portwatch %s <--\n",VERSION_NO);
	printf("By Mark Kamichoff\n\n");

//	if(!((argc != 2) || (argc != 3)))
	if( (argc != 2) && (argc != 3) )
	{
		printf("Parse error: Syntax is %s portnum [-l]\n",
			argv[0]);
		exit(1);
	}

	if((argc == 3) && !strcmp(argv[2],"-l") )
		logging = 1;

	if(!((atoi(argv[1]) < 65537) && (atoi(argv[1]) > 0 )))
	{
		fputs("Error: Port number not valid, ",stderr);
		fputs("needs to be between 0 and 65536\n",stderr);
		exit(1);
	}

	/* Gets info about your local UNIX machine */

	if(uname(&myname) == -1)
	{
		printf("Error getting system identification, ");
		printf("this can't be good; exiting!\n");
		exit(1);
	}

	printf("System Info: %s, Version %s, Architecture: %s\n",
		myname.sysname,
		myname.release,
		myname.machine);

	/* This will come out something like this:
	System Info: Linux, Version 2.4.17, Architecture: i686
	*/

	printf("Attempting to create socket...");
	if(!(listeningfd = socket(AF_INET, SOCK_STREAM, 0)))
	{
		fputs("Error: Socket creation failed\n",stderr);
		exit(1);
	}
	printf("...success!\n");

			/* Clear contents of servaddr */
	bzero(&servaddr, sizeof(servaddr));

	/* Set the local server address */

	servaddr.sin_family = AF_INET;	/* IP */

	/* Listens on all addresses */
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	/* Sets the port */
	servaddr.sin_port = htons(atoi(argv[1]));

	if(bind(listeningfd, (SA *) &servaddr, sizeof(servaddr)) != 0)
	{
		fputs("Error: Bind() error, ",stderr);
		fputs("the port is probably in use\n",stderr);

		if(atoi(argv[1]) < 1024) {
			fputs("Note: You need root for ports < 1024\n",stderr);
		}

		exit(1);
	}

	/* Set the socket to listen for a connection */

	if(listen(listeningfd, 5) != 0)
	{
		fputs("Error: Call to listen() failed, ", stderr);
		fputs("perhaps a socket error\n",stderr);
		exit(1);
	}

	/* Just saying obvious stuff... */
	printf("Listening on port %s of all local interfaces...\n",argv[1]);

	while(1)	/* Keeps accepting connections until termination */
	{
		clientlength = sizeof(clientaddress);
		clientfd = accept(listeningfd, (SA *) &clientaddress,
			&clientlength);

		hptr = gethostbyaddr((struct in_addr *) &clientaddress.sin_addr,
			sizeof(clientaddress.sin_addr),
			AF_INET);

		if(fork() == 0)
		{
			close(listeningfd);

			if(hptr != NULL)	/* If we looked up hostname */
			{
				/* Get Current Time */
				ticks = time(NULL);

				printf("Connect: %s:%d [%s] at %.24s\r\n",
					hptr->h_name,
					ntohs(clientaddress.sin_port),
					inet_ntoa(clientaddress.sin_addr),
					ctime(&ticks));
			}
			else			/* If query NAK'ed */
			{
				/* Get Current Time */
				ticks = time(NULL);

				printf("Connect: %s:%d [%s] at %.24s\r\n",
					inet_ntoa(clientaddress.sin_addr),
					ntohs(clientaddress.sin_port),
					inet_ntoa(clientaddress.sin_addr),
					ctime(&ticks));
			}

			if(logging == 1)
			{
				if(hptr != NULL)
				{
					keeplog(clientfd,
						(char *)
						inet_ntoa(clientaddress.sin_addr),
						(unsigned long)
						ntohs(clientaddress.sin_port),
						(char *) hptr->h_name);
				}
				else
				{
					keeplog(clientfd,
						(char *)
						inet_ntoa(clientaddress.sin_addr),
						(unsigned long)
						ntohs(clientaddress.sin_port),
						(char *)
						inet_ntoa(clientaddress.sin_addr));
				}
			}
			else
			{
				while((n = read(clientfd, junkbuffer,
					strlen(junkbuffer))) != 0);

				/* Get Current Time */
				ticks = time(NULL);

				printf("Disconnect: %s:%d [%s] at %.24s\r\n",
					hptr->h_name,
					ntohs(clientaddress.sin_port),
					inet_ntoa(clientaddress.sin_addr),
					ctime(&ticks));
			}

			exit(0);
		}
		close(clientfd);
	}

	exit(0);
}
